// -----------------------------------------------------------------------------
// |    ___  ___  _  _ _     ___                                        _      |
// |   / __>| . \| || \ |   | __>_ _  ___ ._ _ _  ___  _ _ _  ___  _ _ | |__   |
// |   \__ \|  _/| ||   |   | _>| '_><_> || ' ' |/ ._>| | | |/ . \| '_>| / /   |
// |   <___/|_|  |_||_\_|   |_| |_|  <___||_|_|_|\___.|__/_/ \___/|_|  |_\_\   |
// |                                                                           |
// |---------------------------------------------------------------------------|
//
// http://spinframework.sourceforge.net
// Copyright (C) 2009 Mike Wozniewski, Zack Settel
//
// Developed/Maintained by:
//    Mike Wozniewski (http://www.mikewoz.com)
//    Zack Settel (http://www.sheefa.net/zack)
//
// Principle Partners:
//    Shared Reality Lab, McGill University (http://www.cim.mcgill.ca/sre)
//    La Societe des Arts Technologiques (http://www.sat.qc.ca)
//
// Funding by:
//    NSERC/Canada Council for the Arts - New Media Initiative
//    Heritage Canada
//    Ministere du Developpement economique, de l'Innovation et de l'Exportation
//
// -----------------------------------------------------------------------------
//  This file is part of the SPIN Framework.
//
//  SPIN Framework is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  SPIN Framework is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with SPIN Framework. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------

#ifndef __SceneManager_H
#define __SceneManager_H

#include <osg/Node>
#include <osg/Group>
#include <osg/ClearNode>
#include <osg/Geode>

#include <osgDB/SharedStateManager>
#include <osgShadow/SoftShadowMap>
#include <cppintrospection/Value>

#include "config.h"
#include "referencednode.h"
#include "referencedstateset.h"

#include "usernode.h"

#include "lo/lo.h"
#include "tinyxml.h"

#ifdef WITH_BULLET
#include <LinearMath/btVector3.h>
class btCollisionWorld;
class btDynamicsWorld;
#endif

namespace spin
{


// forward declarations:
class GroupNode;
class UserNode;
class spinLog;

/**
 * \brief The main class that maintains the scene and handles OSC messages.
 *
 * The SceneManager class should only be instantiated once for each process
 * or thread. However, this should rarely be done directly because the
 * spinContext instance (listener or server) have a better interface
 * for including SPIN in an application.
*/
class SceneManager
{
    public:
        //SceneManager(const std::string &id);
        SceneManager(std::string id);
        ~SceneManager();

        void init();
        void debugContext();
        void debugNodes();
        void debugStateSets();
        void debugSceneGraph();
        void debug();

        void setGraphical(bool b) { graphicalMode = b; }
        bool isGraphical() { return (bool) graphicalMode; }

        //void setLogFile(const char *logfile);
        void setLog(spinLog& log);

        void registerStateSet(ReferencedStateSet *s);
        void unregisterStateSet(ReferencedStateSet *s);

        void sendNodeTypes(lo_address txAddr = 0);
        void sendStateTypes(lo_address txAddr = 0);
        void sendNodeList(std::string type, lo_address txAddr = 0);
        void sendConnectionList(lo_address txAddr = 0);

        std::vector<std::string> getAllNodeTypes();

        ReferencedNode *createNode(std::string id, std::string type);
        ReferencedNode *createNode(const char *id, const char *type);
        ReferencedNode *getNode(std::string id);
        ReferencedNode *getNode(const char *id);
        ReferencedNode *getNode(const char *id, const char *type);
        ReferencedNode *getOrCreateNode(const char *id, const char *type);

        ReferencedStateSet* getStateSet(const char *id);
        ReferencedStateSet* createStateSet(const char *id, const char *type);
        ReferencedStateSet* createStateSet(const char *fname);
        
        void setWorldStateSet(const char *s);

        std::vector<ReferencedNode*> findNodes(const char *pattern);
        std::vector<ReferencedStateSet*> findStateSets(const char *pattern);

        /**
         * This method removes a node from the scene, however the actual work is
         * done by the doDelete() method.
         *
         * NOTE: All children of the node will remain, but will be re-attached
         * to the 'world' node instead. If you want to destroy all children as
         * well, then use deleteGraph().
         */
        void deleteNode(const char *id);

        /**
         * deleteGraph() operates similarly to deleteNode(), except that all
         * children are also deleted.
         */
        void deleteGraph(const char *id);

        /**
         * The doDelete method performs all of the necessary steps to remove a
         * node from the scene: The node's detach() method is called, which will
         * actually remove it from the scene graph, and eliminate the OSC
         * callback. The callbackUpdate() function is unregistered. And finally,
         * a message is broadcasted to all clients.
         */
        void doDelete(ReferencedNode *n);

        /**
         * The doDelete method performs all of the necessary steps to delete a
         * stateset. It is similar to the doDelete method for a note: it calls
         * the node's removeFromScene() method, releases resources (eg, videos),
         * removes it from the stateMap list, etc.
         */
        void doDelete(ReferencedStateSet *n);
        
        /**
         * Clears scene elements that are not part of any user's subgraphs
         */
        void clear();

        /**
         * Clears only the users from the scene (and any attached nodes in their
         * subgraphs
         */
        void clearUsers();

        /**
         * Forces a removal of all states from the scene graph. This should only
         * be used to clean up upon exit.
         */
        void clearStates();

        /**
         * The update method is where any thread-safe changes to the scene graph
         * should go. The method is guaranteed to be called only when there are
         * no traversals being performed.
         */
        void update();

        osg::Matrix getWorldCoords(t_symbol *id);

        void exportScene (const char *nodeID, const char *filename);

        std::string sceneID;

        osg::ref_ptr<osg::ClearNode> rootNode;
        //osg::ref_ptr<osg::ClearNode> worldNode;
        osg::ref_ptr<ReferencedNode> worldNode;
        osg::ref_ptr<osg::Geode> gridGeode;
        
        t_symbol* worldStateSet_;

        /**
         * The scene manager can operate in graphical mode or non-graphical.
         * > For graphical mode, the full scene graph structure is instantiated,
         *   and can thus be renderered by an OSG rendering process.
         * > Non-graphical mode only maintains the basic features in the scene
         *   graph that are needed for GUIs and information servers. Eg, lights
         *   are not created.
         */
        bool graphicalMode;

        osg::ref_ptr<GroupNode> globalObserver;

        bool activeLights[OSG_NUM_LIGHTS];

        std::string getStateAsXML(std::vector<lo_message> nodeState);
        std::string getNodeAsXML(ReferencedNode *n, bool withUsers, bool withChildren);

        std::string getConnectionsAsXML();
        std::vector<t_symbol*> getSavableStateSets(ReferencedNode *n, bool withUsers);

        /**
         * Write the entire scene to an XML file. If no absolute path is
         * specified, the file will be written to the ~/.spinframework folder.
         */
        bool saveXML(const char *filename, bool withUsers = false);
        
        /**
         * Write exactly one node to an XML file, without it's subgraph
         * (ie, children nodes) or statesets.
         */
        bool saveNode(const char *nodeID, const char *filename);
        
        /**
         * Save all the UserNodes in a scene to an XML file.
         */
        bool saveUsers(const char *filename);

        /**
         * Loads any nodes or statesets from a file, and sets the parent node
         */
        bool createNodeFromXML(TiXmlElement *XMLnode, const char *parentNode);
        bool createStateSetFromXML(TiXmlElement *XMLnode);
        bool createConnectionsFromXML(TiXmlElement *XMLnode);
        bool loadXML(const char *filename);

        osg::ref_ptr<osgDB::SharedStateManager> sharedStateManager;

        /**
         * The refreshAll method results in a broadcast of all nodelists so that
         * clients can create any missing nodes. Then, the full node state is
         * broadcasted, for ALL nodes.
         */
        void refreshAll();
        
        /**
         * The refreshSubscribers method results in a publication of all nodelists to the given TCP 
         * subscribers. Then, the full node state is
         * published to the subscribers, for ALL nodes.
         */
        void refreshSubscribers(const std::map<std::string, lo_address> &clients);

        /**
         * Enable the shadow system
         **/
        void setShadows(bool b);

        /**
         * Set the values for width of the soft penumbra the shader will use.
         */
        void setShadowSoftness(float f);
        /**
         * Set the values for jittering scale the shader will use. 
         */
        void setShadowJitter(float f);
        /**
         * Add a small bias to the z-value, this can reduce shadow acne problem. 
         */
        void setShadowBias(float f);
        /**
         * The ambiantBias defines how much shadows darken the scene. It is
         * usually used to set up lower bound for shadowing factor, in the range
         * [0..1]. Sometimes we want to make shadows to be only a bit darker
         * than lit areas. This is called "AmbientBias" because shadowed areas
         * are lit only by the ambient light component). AmbientBias.y is
         * usually set up to 1 - AmbientBias.x but it may be also set bit larger
         * or smaller values used to make shadow range more dynamic or flat. 
         */
        void setShadowAmbientBias(float x, float y);

        /**
         * Set the gravity vector for the physics engine (only used by some
         * nodes).
         */
        void setGravity(float x, float y, float z);
        
        void setWind(float x, float y, float z);

        /**
         * Sets the update delay for the physics engine (in seconds). The
         * default is 0.02 seconds (ie, 50hz).
         */
        void setUpdateRate(float seconds);
        

#ifdef WITH_BULLET
        bool lastColState;
        void detectCollision( bool& lastColState, btCollisionWorld* cw );
        btDynamicsWorld *dynamicsWorld_;
        btVector3 wind_;
        bool windy_;
#endif

    private:

        osg::ref_ptr<osgShadow::ShadowedScene> shadowRoot_;
        osg::ref_ptr<osgShadow::SoftShadowMap> softShadowMap_;

        static bool nodeSortFunction (osg::ref_ptr<ReferencedNode> n1, osg::ref_ptr<ReferencedNode> n2);
        //std::vector< osg::ref_ptr<ReferencedNode> > nodeList;
        nodeMapType nodeMap; // the nodeList arranged by type
        ReferencedStateSetMap stateMap;
        
        

        float dynamicsUpdateRate_; // in seconds
        osg::Timer_t lastTick_;
};

/**
 * \namespace introspector
 * \brief Function that deal with cppintrospection-wrapped namespaces, classes and methods from SPIN.
 */
namespace introspector
{
    /**
     * Prepends "spin::" to a type name.
     */
    std::string prependNamespace(const std::string &name);
    /**
     * Prepends "spin::" to a type name and returns its cppintrospection::Type.
     */
    const cppintrospection::Type& getType(const std::string &name);

    /**
     * Recursive function to invoke a method for a particular class, that will try
     * all base classes as well
     * 
     * This OSC callback functions need to be valid C function pointers, so it's
     * declared here as a function, not a method):
     */
    int invokeMethod(const cppintrospection::Value classInstance, const cppintrospection::Type &classType, std::string method, cppintrospection::ValueList theArgs);

#if 0
    /**
     * Removes "spin::" from a type name.
     */
    std::string removeNamespace(const std::string &name);
#endif

} // end of namespace introspector


} // end of namespace spin

#endif
